/* hypercall.S - wrappers for Xen hypercalls */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2013  Free Software Foundation, Inc.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <grub/symbol.h>
#include <grub/xen.h>

FUNCTION(grub_xen_hypercall)
	pushl   %ebp
	movl    %esp, %ebp
	pushl   %esi
	pushl   %edi
	pushl   %ebx

	/* call number already in %eax. */
	/* %edx -> %ebx*/
	/* %ecx -> %ecx*/
	movl   %edx,  %ebx
	movl    8(%ebp), %edx
	movl    12(%ebp), %esi
	movl    16(%ebp), %edi
	movl    20(%ebp), %ebp
	int     $0x82
	popl   %ebx
	popl   %edi
	popl   %esi
	popl   %ebp
	ret
