/^Project-Id-Version:/ b
/^Report-Msgid-Bugs-To:/ b
/^POT-Creation-Date:/ b
/^PO-Revision-Date:/ b
/^Last-Translator:/ b
/^Language-Team:/ b
/^Language:/ b
/^MIME-Version:/ b
/^Content-Type:/ b
/^Content-Transfer-Encoding:/ b
/^Plural-Forms:/ b

s,a,а,g
s,b,б,g
s,c,ц,g
s,d,д,g
s,e,е,g
s,f,ф,g
s,g,г,g
s,h,х,g
s,i,и,g
s,j,й,g
s,k,к,g
s,l,л,g
s,m,м,g
s,n,н,g
s,o,о,g
s,p,п,g
s,q,ќ,g
s,r,р,g
s,s,с,g
s,t,т,g
s,u,у,g
s,v,в,g
s,w,ў,g
s,x,ѯ,g
s,y,ѝ,g
s,z,з,g
s,A,А,g
s,B,Б,g
s,C,Ц,g
s,D,Д,g
s,E,Е,g
s,F,Ф,g
s,G,Г,g
s,H,Х,g
s,I,И,g
s,J,Й,g
s,K,К,g
s,L,Л,g
s,M,М,g
s,N,Н,g
s,O,О,g
s,P,П,g
s,Q,Ќ,g
s,R,Р,g
s,S,С,g
s,T,Т,g
s,U,У,g
s,V,В,g
s,W,Ў,g
s,X,Ѯ,g
s,Y,Ѝ,g
s,Z,З,g
s,А,А,g
s,Б,Б,g
s,Ц,Ц,g
s,Д,Д,g
s,Е,Е,g
s,Ф,Ф,g
s,Г,Г,g
s,Х,Х,g
s,И,И,g
s,Й,Й,g
s,К,К,g
s,Л,Л,g
s,М,М,g
s,Н,Н,g
s,О,О,g
s,П,П,g
s,Ќ,Ќ,g
s,Р,Р,g
s,С,С,g
s,Т,Т,g
s,У,У,g
s,В,В,g
s,Ў,Ў,g
s,Ѯ,Ѯ,g
s,Ѝ,Ѝ,g
s,З,З,g

s,%\([0-9]*\)Ц,%\1C,g
s,%\([0-9]*\)ц,%\1c,g
s,%\([0-9]*\)д,%\1d,g
s,%\([0-9]*\)лд,%\1ld,g
s,%\([0-9]*\)ллд,%\1lld,g

s,%\([0-9\.\*]*\)с,%\1s,g
s,%\([0-9\.\*]*\)м,%\1m,g
s,%\([0-9]*\)у,%\1u,g
s,%\([0-9]*\)лу,%\1lu,g
s,%\([0-9]*\)ллу,%\1llu,g
s,%\([0-9]*\)зу,%\1zu,g
s,%\([0-9]*\)ѯ,%\1x,g
s,%\([0-9]*\)лѯ,%\1lx,g
s,%\([0-9]*\)ллѯ,%\1llx,g
